/*
 * ProfilerInformation.h - profile information header file.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifndef PROFILERINFORMATION_H
#define PROFILERINFORMATION_H

#include <utility>
#include <list>
#include <vector>
#include <map>
#include <set>
#include <string>

#include <cstddef>

// #include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"

class SMPProgram;	// forward declaration so we can declare a pointer to an SMPProgram
class MemoryAccessInfo;

class InstructionInformation
{
	public:
		InstructionInformation() : numeric_count(0), pointer_count(0), other_count(0) {};
		virtual ~InstructionInformation() {};

		inline int isNumeric() { return (numeric_count > 0) && (pointer_count == 0) && (other_count == 0); }
		inline void setNumericCount(int nc) { numeric_count = nc;}
		inline void setPointerCount(int pc) { pointer_count = pc;}
		inline void setOtherCount(int oc) { other_count = oc;}
		inline long long getNumericCount() { return numeric_count; }
		inline long long getPointerCount() { return pointer_count; }
		inline long long getOtherCount() { return other_count; }

	private:

		long long numeric_count;
		long long pointer_count;
		long long other_count;
};

class ProfilerInformation
{
	public:
		ProfilerInformation(const char *, SMPProgram *);
		virtual ~ProfilerInformation();

		inline InstructionInformation* GetInfo(ea_t addr) { return the_map[addr]; };

		std::set<ea_t> GetIndirectCallTargets(ea_t addr) { return indirect_call_map[addr]; };
		inline MemoryAccessInfo *GetMemoryAccessInfo(void) { return MemProfileInfo; };
		inline long GetProfilerAnnotationCount(void) const { return ProfilerAnnotsRead; };

	protected:
		// add type information from profiler about loads
		void addProfileInformation(ea_t addr, long long nc, long long pc, long long oc);
		
		// add pre-dec info from profiler
		void profileAddressingInformation ( ea_t addr, int size, char *type, 
			char *scope, ea_t the_const, char* field, ea_t real_const);

		inline void SetInfo(ea_t addr, InstructionInformation* ii) { the_map[addr] = ii; }

	private:

		// annotations that are used for numeric type propagation.
		std::map<ea_t,InstructionInformation*> 	the_map;	
	
		// annotations to spew back out verbatim.
		std::list<std::string>			constant_info;	

		std::string filename;

		std::map<ea_t, std::set<ea_t> > indirect_call_map;

		MemoryAccessInfo *MemProfileInfo;
		SMPProgram *MyProg;

		// Count of profiler annotations read in.
		long ProfilerAnnotsRead;
};

// Ways in which a memory byte can be accessed.
enum SMPAccessType {
	SMP_NO_ACCESS = 0, // lattice top; uninitialized
	SMP_DIRECT = 1,  // not indirect, not scaled
	SMP_SCALED = 2,  // indirect with scaled index register used
	SMP_INDIRECT = 4 // indirect, but not scaled
};

#define SMP_NO_REGION 0
#define SMP_GLOBAL 1
#define SMP_STACK 2
#define SMP_HEAP 4

// A class to hold profiler information about the memory accesses performed
//  at an instruction address.
class SMPMemoryAccesses {
public:
	// Constructors and destructors.
	SMPMemoryAccesses(ea_t addr);

	// Get functions.
	inline ea_t GetAddr(void) const { return address; };
	inline size_t GetLimit(void) const { return ByteLimit; };
	inline SMPAccessType GetAccessType(size_t index) const {
		return ProfTypes.at(index);
	}
	inline int GetRegion(void) const { return region; };

	// Set functions.
	inline void SetLimit(size_t NewLimit) { 
		ByteLimit = NewLimit; 
		ProfTypes.assign(NewLimit, SMP_NO_ACCESS);
		return; 
	};
	inline void SetAccessType(size_t index, SMPAccessType TypeSeen) {
		ProfTypes[index] = TypeSeen;
		return;
	}
	inline void SetRegion(int NewRegion) {
		region |= NewRegion;
		return;
	}

	// Analysis functions.
	void MeetAccessType(size_t index, SMPAccessType TypeSeen);

private:
	ea_t address;
	size_t ByteLimit;  // number of bytes in data object
	int region;
	std::vector<SMPAccessType> ProfTypes;  // access types seen by profiler,
						// indexed by byte offset from start of object
}; // end class SMPMemoryAccesses

// A class to summarize profiled memory access granularity information across
//  the entire executable.
class MemoryAccessInfo {
public:
	// Contructors and destructors.
	MemoryAccessInfo(SMPProgram *CurrProg);
	virtual ~MemoryAccessInfo();
	// Debug methods.
	void Dump(void);
	// Analysis methods.
	void ProcessMemoryAccessAnnotation(FILE *fin, int addr);
	void InferDataGranularity(void);
	bool ComputeNonDirectAccessRegion(ea_t InstAddr, int &ChildOffset,
		int &RegionSize); // If true is returned, InstAddr makes non-direct
						// accesses to child object from ChildOffset to
						// ChildOffset + RegionSize - 1 within parent object.

private:
	SMPProgram *MyProg; // SMPProgram object for executable

	// The set of instruction addresses that access objects of multiple sizes,
	//  which implies that the instructions access multiple types (e.g. the
	//  copying or initializing instructions in memset(), memcpy(), and the
	//  customized routines that perform similar functions in an application).
	std::set<ea_t> MultipleAccessTypeSet;

	// Map of instruction addresses to the accesses performed by them.
	std::map<ea_t, SMPMemoryAccesses *> InstrAccessMap;

	// Map of instruction addresses to the object allocation addresses for
	//  all objects accessed by that instruction.
	// NOTE: For static data, the "allocation address" is the data address.
	//  For heap and stack data, the allocation address is truly the code
	//  address at which the data is allocated.
	std::map<ea_t, std::set<ea_t> > InstrObjectMap;

	// Inverse mapping: From object allocation addresses to all instruction
	//  addresses that access the object.
	std::map<ea_t, std::set<ea_t> > ObjectInstrMap;

	// Map of memory access instruction addresses to the data type number
	//  for the data objects accessed by that code address.
	std::map<ea_t, size_t> InstrTypeMap;
	// Ditto for mapping object allocation addresses to data type ID #.
	std::map<ea_t, size_t> ObjectTypeMap;

	// Meet of access types for each data object type, with outer index
	//  being the type number from InstrTypeMap and inner index being
	//  the byte offset within the data object.
	std::vector<SMPMemoryAccesses *> TypeAccessMap;

	// Method to recursively associate a data type ID with all instruction
	//  addresses that access objects of that data type.
	void AssociateInstrWithType(ea_t InstAddr, size_t TypeID,
		std::map<ea_t, std::set<ea_t> >::iterator InstObjIter);
}; // end of class MemoryAccessInfo

#endif

