/*
 * SMPBasicBlock.h - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifndef SMPBASICBLOCK_H
#define SMPBASICBLOCK_H 1

// SMPBasicBlock.h
//
// This header defines the interfaces needed for analyzing basic blocks.

#include <list>
#include <set>
#include <string>

#include <cstddef>

#include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"
#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPInstr.h"

// We can use less memory if we don't build explicit def-use chains.
#define STARS_BUILD_DEF_USE_CHAINS 0

// Aggressively propagate stack memory types until we see an indirect memory write
#define STARS_PROPAGATE_STACKMEMTYPE_UNTIL_INDIRECT 1

// Value for a block number before it is initialized.
#define SMP_BLOCKNUM_UNINIT (-1)

// Value for an SSA subscript number before it is initialized by SSA renaming.
#define SMP_SSA_UNINIT  (-1)

// Masks to set bits to use as booleans1 in SMPBasicBlock.
#define BLOCK_SET_PROCESSED 0x01
#define BLOCK_SET_INDIRECT_JUMP 0x02
#define BLOCK_SET_RETURNS 0x04
#define BLOCK_SET_SHARED_TAIL_CHUNK 0x08
#define BLOCK_SET_LIVE_IN_STALE 0x10
#define BLOCK_SET_UNREACHABLE_BLOCK 0x20
#define BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL 0x40
#define BLOCK_SET_ALIASED_WRITE 0x80
#define BLOCK_RESET_PROCESSED 0xfe
#define BLOCK_RESET_INDIRECT_JUMP 0xfd
#define BLOCK_RESET_RETURNS 0xfb
#define BLOCK_RESET_SHARED_TAIL_CHUNK 0xf7
#define BLOCK_RESET_LIVE_IN_STALE 0xef
#define BLOCK_RESET_UNREACHABLE_BLOCK 0xdf
#define BLOCK_RESET_FLAGS_DEAD_BEFORE_FIRST_KILL 0xbf
#define BLOCK_RESET_ALIASED_WRITE 0x7f

// Masks to set bits to use as booleans2 in SMPBasicBlock.
#define BLOCK_SET_REACHES_OUT_STALE 0x01
#define BLOCK_SET_LOOP_TAIL_BLOCK 0x02
#define BLOCK_SET_LOOP_HEADER_BLOCK 0x04
#define BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC 0x08
#define BLOCK_SET_SCCP_VISITED 0x10
#define BLOCK_SET_HAS_CALL_INST 0x20
#define BLOCK_SET_UNUSED7 0x40
#define BLOCK_SET_UNUSED8 0x80
#define BLOCK_RESET_REACHES_OUT_STALE 0xfe
#define BLOCK_RESET_LOOP_TAIL_BLOCK 0xfd
#define BLOCK_RESET_LOOP_HEADER_BLOCK 0xfb
#define BLOCK_RESET_CALLS_UNSIGNED_ARG_FUNC 0xf7
#define BLOCK_RESET_SCCP_VISITED 0xef
#define BLOCK_RESET_HAS_CALL_INST 0xdf
#define BLOCK_RESET_UNUSED7 0xbf
#define BLOCK_RESET_UNUSED8 0x7f

// Class defining basic blocks.
class SMPBasicBlock {
public:
	// Constructors
	SMPBasicBlock(SMPFunction *Func, std::list<SMPInstr *>::iterator FirstInstr, std::list<SMPInstr *>::iterator LastInstr);

	// Operators
	int operator==(const SMPBasicBlock &rhs) const;

	// Get methods
	ea_t GetFirstAddr(void) const;
	ea_t GetLastAddr(void);
	inline int GetNumber(void) const { return BlockNum; };
	inline SMPBasicBlock *GetThisBlock(void) { return this; };
	inline SMPFunction *GetFunc(void) { return MyFunc; };
#if 0
	inline sval_t GetIncomingStackPtrDelta(void) const { return IncomingStackPtrDelta; };
#endif
	inline std::vector<SMPInstr *>::iterator GetFirstInst(void) { return InstVec.begin(); };
	inline std::vector<SMPInstr *>::iterator GetLastInst(void) { return InstVec.end(); };
	inline std::vector<SMPInstr *>::reverse_iterator GetRevInstBegin(void) { return InstVec.rbegin(); };
	inline std::vector<SMPInstr *>::reverse_iterator GetRevInstEnd(void) { return InstVec.rend(); }
	inline std::list<SMPBasicBlock *>::iterator GetFirstPred(void) {
		return Predecessors.begin();
	};
	inline std::list<SMPBasicBlock *>::iterator GetLastPred(void) {
		return Predecessors.end();
	};
	inline std::list<SMPBasicBlock *>::iterator GetFirstSucc(void) {
		return Successors.begin();
	};
	inline std::list<SMPBasicBlock *>::iterator GetLastSucc(void) {
		return Successors.end();
	};
	inline std::list<SMPBasicBlock *>::reverse_iterator GetRevSuccBegin(void) {
		return Successors.rbegin();
	};
	inline std::list<SMPBasicBlock *>::reverse_iterator GetRevSuccEnd(void) {
		return Successors.rend();
	};
	std::list<SMPBasicBlock *>::iterator GetFallThroughSucc(void); // Search successors for fall-through block, if any
	inline size_t GetNumPreds(void) const { return Predecessors.size(); };
	inline size_t GetNumSuccessors(void) const { return Successors.size(); };
	inline int GetLoopHeaderNumber(void) const { return LoopHeaderNumber; };
	size_t GetBlockSize(void); // returns size in bytes
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLiveIn(void); // LiveInSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLiveIn(void); // LiveInSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLiveOut(void); // LiveOutSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLiveOut(void); // LiveOutSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstVarKill(void); // KillSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastVarKill(void); // KillSet.end()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstUpExposed(void); // UpExposedSet.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastUpExposed(void); // UpExposedSet.end()
	inline size_t GetUpExposedSetSize(void) const { return UpExposedSet.size(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetFirstReachesIn(void) { return ReachesInSet.begin(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetLastReachesIn(void) { return ReachesInSet.end(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetFirstReachesOut(void) { return ReachesOutSet.begin(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetLastReachesOut(void) { return ReachesOutSet.end(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetFirstDownExposedDefn(void) { return DownExposedDefnSet.begin(); };
	inline std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition>::iterator GetLastDownExposedDefn(void) { return DownExposedDefnSet.end(); };
	std::set<STARSOpndTypePtr, LessOp>::iterator GetFirstLocalName(void); // LocalNames.begin()
	std::set<STARSOpndTypePtr, LessOp>::iterator GetLastLocalName(void); // LocalNames.end()
	std::set<int>::iterator GetFirstDomFrontier(void); // DomFrontier.begin()
	std::set<int>::iterator GetLastDomFrontier(void); // DomFrontier.end()
	std::set<SMPPhiFunction, LessPhi>::iterator GetFirstPhi(void); // PhiFunctions.begin()
	std::set<SMPPhiFunction, LessPhi>::iterator GetLastPhi(void); // PhiFunctions.end()
	int GetPredPosition(int BlockNum); // What is position # within Preds of BlockNum?
	std::set<SMPPhiFunction, LessPhi>::iterator FindPhi(STARSOpndTypePtr FindOp);
	std::set<STARSOpndTypePtr, LessOp>::iterator FindLocalName(STARSOpndTypePtr FindOp);
	inline std::map<int, struct STARS_SCCP_Const_Struct>::iterator FindLocalConstValue(int DefHashValue) { return ConstantLocalDefs.find(DefHashValue); };
	inline std::map<int, struct STARS_SCCP_Const_Struct>::iterator GetLastLocalConstValueIter(void) { return ConstantLocalDefs.end(); };
#if 0
	// Given a USE operand and the address of its instr, return DEF from DU chains or Phi func.
	std::set<DefOrUse, LessDefUse>::iterator GetDefFromOpAddr(STARSOpndTypePtr UseOp, ea_t InstAddr, int SSANum);
#endif
	std::set<DefOrUse, LessDefUse>::iterator GetGlobalDefIterFromDefAddr(STARSOpndTypePtr DefOp, ea_t DefAddr);
	ea_t GetDefAddrFromUseAddr(STARSOpndTypePtr UseOp, ea_t InstAddr, int SSANum, bool LocalName); // return DefAddr, block # for Phi Defs, or UpExposed address.
	ea_t GetUltimateDefAddr(STARSOpndTypePtr UseOp, ea_t InstAddr, int SSANum, bool LocalName); // trace into predecessors if necessary

	SMPInstr *FindInstr(ea_t InstAddr);
	std::vector<SMPInstr *>::iterator GetInstIterFromAddr(ea_t InstAddr);
	inline SMPInstr *GetInstFromIndex(size_t VecIndex) const { return InstVec.at(VecIndex); };
	size_t GetIndexFromInstAddr(ea_t InstAddr) const; // Get the InstVec index for the instruction at InstAddr. Assert if not found.
	inline sval_t GetOutgoingStackDelta(void) const { return OutgoingStackDelta; };

	// Six methods to get values from the maps of local reg/SSA to FG info.
	//  For global names, see corresponding methods in SMPFunction.
	unsigned short GetDefSignMiscInfo(int DefHashValue);
	unsigned short GetUseSignMiscInfo(int UseHashValue);
	unsigned short GetDefWidthTypeInfo(int DefHashValue);
	unsigned short GetUseWidthTypeInfo(int UseHashValue);
	struct FineGrainedInfo GetDefFGInfo(int DefHashValue);
	struct FineGrainedInfo GetUseFGInfo(int UseHashValue);

	// Set methods
#if 0
	inline void SetIncomingStackPtrDelta(sval_t delta) { IncomingStackPtrDelta = delta; };
#endif
	inline void SetProcessed(bool flag) { flag ? (booleans1 |= BLOCK_SET_PROCESSED) : (booleans1 &= BLOCK_RESET_PROCESSED); };
	inline void SetMaybeAliased(bool flag) { flag ? (booleans1 |= BLOCK_SET_ALIASED_WRITE) : (booleans1 &= BLOCK_RESET_ALIASED_WRITE); };
	inline void SetIndirectJump(bool flag) { flag ? (booleans1 |= BLOCK_SET_INDIRECT_JUMP) : (booleans1 &= BLOCK_RESET_INDIRECT_JUMP); };
	inline void SetReturns(bool flag) { flag ? (booleans1 |= BLOCK_SET_RETURNS) : (booleans1 &= BLOCK_RESET_RETURNS); };
	inline void SetShared(void) { booleans1 |= BLOCK_SET_SHARED_TAIL_CHUNK; };
	inline void ResetShared(void) { booleans1 &= BLOCK_RESET_SHARED_TAIL_CHUNK; };
	inline void SetLiveInStale(bool flag) { flag ? (booleans1 |= BLOCK_SET_LIVE_IN_STALE) : (booleans1 &= BLOCK_RESET_LIVE_IN_STALE); };
	inline void SetUnreachableBlock(bool flag) { flag ? (booleans1 |= BLOCK_SET_UNREACHABLE_BLOCK) : (booleans1 &= BLOCK_RESET_UNREACHABLE_BLOCK); };
	inline void SetFlagsDeadBeforeFirstKill(bool flag) { flag ? (booleans1 |= BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL) : (booleans1 &= BLOCK_RESET_FLAGS_DEAD_BEFORE_FIRST_KILL); };
	inline void SetReachesOutStale(void) { booleans2 |= BLOCK_SET_REACHES_OUT_STALE; };
	inline void SetLoopHeaderBlock(void) { booleans2 |= BLOCK_SET_LOOP_HEADER_BLOCK; };
	inline void ResetLoopHeaderBlock(void) { booleans2 &= BLOCK_RESET_LOOP_HEADER_BLOCK; };
	inline void ResetLoopTailBlock(void) { booleans2 &= BLOCK_RESET_LOOP_TAIL_BLOCK; };
	inline void SetCallsUnsignedArgFunc(void) { booleans2 |= BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC; };
	inline void SetSCCPVisited(bool flag) { flag ? (booleans2 |= BLOCK_SET_SCCP_VISITED) : (booleans2 &= BLOCK_RESET_SCCP_VISITED); };
	inline void SetHasCallInst(void) { booleans2 |= BLOCK_SET_HAS_CALL_INST; };
	inline void SetNumber(int Num) { BlockNum = Num; };
	inline void SetOutgoingStackDelta(sval_t OutDelta) { OutgoingStackDelta = OutDelta; };
	void LinkToPred(SMPBasicBlock *Predecessor);
	void LinkToSucc(SMPBasicBlock *Successor);
	inline void AddLiveIn(STARSOpndTypePtr LiveOp) { LiveInSet.insert(LiveOp); };
	inline void AddLiveOut(STARSOpndTypePtr LiveOp) { LiveOutSet.insert(LiveOp); };
	inline void AddVarKill(STARSOpndTypePtr KillOp) { KillSet.insert(KillOp); };
	inline void AddUpExposed(STARSOpndTypePtr UseOp) { UpExposedSet.insert(UseOp); };
	void AddReachesIn(std::pair<STARSOpndTypePtr, ea_t> ReachInDefn);
	void AddReachesOut(std::pair<STARSOpndTypePtr, ea_t> ReachOutDefn);
	inline void AddDownExposedDefn(std::pair<STARSOpndTypePtr, ea_t> DEDefn) { DownExposedDefnSet.insert(DEDefn); };
	inline void ClearPhiFunctions(void) { PhiFunctions.clear(); };
	bool ErasePhi(STARSOpndTypePtr OldOp); // erase() from phi function list
	void ErasePred(ea_t FirstBlockAddr); // erase() block starting at FirstBlockAddr from Preds list
	void EraseSucc(ea_t FirstBlockAddr); // erase() block starting at FirstBlockAddr from Successors list
	bool AddPhi(SMPPhiFunction NewPhi);
	bool SetPhiDefSSA(STARSOpndTypePtr DefOp, int SSANum);
	bool SetPhiUseSSA(STARSOpndTypePtr UseOp, size_t index, int SSANum);
	std::set<SMPPhiFunction, LessPhi>::iterator SetPhiDefType(STARSOpndTypePtr DefOp, SMPOperandType Type);
	bool SetPhiUseType(STARSOpndTypePtr DefOp, size_t index, SMPOperandType Type);
	std::set<SMPPhiFunction, LessPhi>::iterator SetPhiDefMetadata(STARSOpndTypePtr DefOp, SMPMetadataType Status);
	std::map<int, struct STARS_SCCP_Const_Struct>::iterator InsertLocalConstValue(int DefHashValue, struct STARS_SCCP_Const_Struct NewConstEntry); 
		// Insert SCCP value for local name; change if already found.

	// Six methods to set values into the maps of local reg/SSA to FG info.
	//  For global names, see corresponding methods in SMPFunction.
	void UpdateDefSignMiscInfo(int DefHashValue, unsigned short NewInfo);
	void UpdateUseSignMiscInfo(int UseHashValue, unsigned short NewInfo);
	void UpdateDefWidthTypeInfo(int DefHashValue, unsigned short NewInfo);
	void UpdateUseWidthTypeInfo(int UseHashValue, unsigned short NewInfo);
	void UpdateDefFGInfo(int DefHashValue, struct FineGrainedInfo NewFG);
	void UpdateUseFGInfo(int UseHashValue, struct FineGrainedInfo NewFG);

	void ClearDefSignedness(int DefHashValue);

	// Query methods
	inline bool IsProcessed(void) const { return (booleans1 & BLOCK_SET_PROCESSED); };
	inline bool HasIndirectJump(void) const { return (booleans1 & BLOCK_SET_INDIRECT_JUMP); };
	inline bool HasReturn(void) const { return (booleans1 & BLOCK_SET_RETURNS); };
	inline bool IsShared(void) const { return (booleans1 & BLOCK_SET_SHARED_TAIL_CHUNK); };
	inline bool IsLiveInStale(void) const { return (booleans1 & BLOCK_SET_LIVE_IN_STALE); };
	inline bool IsUnreachableBlock(void) const { return (booleans1 & BLOCK_SET_UNREACHABLE_BLOCK); };
#if 0
	inline bool AreFlagsDeadAfterLastUse(void) const { return (booleans1 & BLOCK_SET_FLAGS_DEAD_AFTER_LAST_USE); };
#endif
	inline bool AreFlagsDeadBeforeFirstKill(void) const { return (booleans1 & BLOCK_SET_FLAGS_DEAD_BEFORE_FIRST_KILL); };
	bool AllNops(void); // Are all instructions in the block no-ops?
	inline bool MaybeAliasedWrite(void) const { return (booleans1 & BLOCK_SET_ALIASED_WRITE); };
	inline bool IsReachesOutStale(void) const { return (booleans2 & BLOCK_SET_REACHES_OUT_STALE); };
	inline bool IsLoopTailBlock(void) const { return (booleans2 & BLOCK_SET_LOOP_TAIL_BLOCK); };
	inline bool IsLoopHeaderBlock(void) const { return (booleans2 & BLOCK_SET_LOOP_HEADER_BLOCK); };
	inline bool CallsUnsignedArgFunc(void) const { return (booleans2 & BLOCK_SET_CALLS_UNSIGNED_ARG_FUNC); };
	inline bool IsSCCPVisited(void) const { return (booleans2 & BLOCK_SET_SCCP_VISITED); };
	inline bool HasCallInstruction(void) const { return (booleans2 & BLOCK_SET_HAS_CALL_INST); };
	inline bool IsSelfLoop(void) const { return (IsLoopHeaderBlock() && (GetNumber() == GetLoopHeaderNumber())); };
	bool IsLiveIn(STARSOpndTypePtr CurrOp);
	inline bool IsLiveOut(STARSOpndTypePtr CurrOp) const {
		return (LiveOutSet.end() != LiveOutSet.find(CurrOp));
	}
	inline bool IsUpExposed(STARSOpndTypePtr CurrOp) const {
		return (UpExposedSet.end() != UpExposedSet.find(CurrOp));
	}
	inline bool IsVarKill(STARSOpndTypePtr CurrOp) const {
		return (KillSet.end() != KillSet.find(CurrOp));
	}
	bool MDAlreadyKilled(STARSOpndTypePtr Opnd1) const; // Was STARSOpndTypePtr killed by something already in KillSet?
	inline bool IsReachesIn(std::pair<STARSOpndTypePtr, ea_t> CurrDefn) const {
		return (ReachesInSet.end() != ReachesInSet.find(CurrDefn));
	}
	inline bool IsReachesOut(std::pair<STARSOpndTypePtr, ea_t> CurrDefn) const {
		return (ReachesOutSet.end() != ReachesOutSet.find(CurrDefn));
	}
	inline bool IsDownExposedDefn(std::pair<STARSOpndTypePtr, ea_t> CurrDefn) const {
		return (DownExposedDefnSet.end() != DownExposedDefnSet.find(CurrDefn));
	}
	bool IsLocalName(STARSOpndTypePtr CurrOp) const;
	bool HasIndirJumpInstruction(void); // true if indirect jump instruction is present
	bool HasConditionalBranch(void); // true if (last) instruction is conditional branch
	bool IsOpDestTruncatedWrite(STARSOpndTypePtr DefOp, int DefSSANum, ea_t DefAddr); // Does DefOp get written out in truncated form (lower bits only)?

	// Printing methods
	void Dump(void);

	// Analysis methods
	bool AllPredecessorsNumbered(void);
	bool AllPredecessorsProcessed(void);
	void DepthFirstMark(void); // Depth-first traversal, mark as processed.
	ea_t GetUltimateOperandSource(ea_t UseAddr, STARSOpndTypePtr UseOp, int UseSSANum); // trace through move and Phi defs to some defining operation.
	void Analyze();
	void InitKilledExposed(bool UseFP); // Initialize KilledSet and UpExposedSet
	bool UpdateLiveOut(void); // Iterate once on updating LiveOutSet; return true if changed
	void SetLiveOutSSANums(void); // Set the SSA numbers for the DEFs in the LiveOut set.
	void AddToDomFrontier(int); // Add RPO block number to DomFrontier set.
	void SetLocalNames(void); // Fill the LocalNames member set
	void SSALocalRenumber(void); // Renumber references to local names
	bool FindLoopHeadsAndTails(void); // Determine if block is loop tail, also set loop header flag for dominating back-edge successor; return false otherwise
	bool DoesDefControlLoopFlow(STARSOpndTypePtr DefOp, int DefSSANum, ea_t DefAddr); // Does DefOp/DefSSANum control loop-back or loop-exit branches?
	void FreeReachingDefsMemory(void); // Free memory for reaching defs sets after they are no longer needed
	void FreeSSAMemory(void); // After SSA #s are in DEFs and USEs, free SSA data structures.
	void FreeUnusedMemory4(void); // After loop 4 (type inference) in SMPProgram::Analyze(), free memory
	bool IsGlobalNameUsedLiveIn(STARSOpndTypePtr GlobalOp); // Is GlobalOp used as a LiveIn value anywhere in this block, including as a Phi USE?
	bool IsGlobalRegDead(ea_t InstAddr, STARSOpndTypePtr Operand, unsigned int RegIndex); // Is global reg dead at InstAddr?
	bool IsRegDead(ea_t InstAddr, uint16 RegNo); // Is local reg dead at InstAddr?
	void MarkDeadRegs(void); // Find dead registers for each mmStrata-instrumented instruction
	bool IsDefDead(ea_t DefAddr, STARSOpndTypePtr DefOp); // Does DefOp at DefAddr never get used?
	bool AreFlagsLiveAfterInst(ea_t InstAddr, std::set<int> &LiveFlagRegs); // Return true and produce set if any flags are live out of inst at InstAddr.
	bool IsDefInvolvedInAddition(ea_t DefAddr, STARSOpndTypePtr DefOp, ea_t &AdditionAddr); // true if DefOp at DefAddr is used in or redefined by addition; pass back AdditionAddr
	bool IsUseCopiedAndShiftedRight(STARSOpndTypePtr UseOp, ea_t UseAddr); // Does a copy of UseOp get shifted right in this block?
	bool DoesDefReachBlockEnd(ea_t DefAddr, STARSOpndTypePtr DefOp, int DefSSANum, std::set<int> &NonEscapingRegisterHashes); // Does value of DefOp/DefSSANum reach block end in any DEF at all?
	bool IsDefOnlyUsedAsAddressReg(ea_t DefAddr, STARSOpndTypePtr DefOp, size_t LoopNum, size_t &AddrUseCounter, SMPOperandType &MemType); // DefOp/DefSSANum is only used in LoopNum (outside of DefAddr) as address reg or Phi USE
	bool IsDefUsedInLoopCompareAndBranch(ea_t DefAddr, STARSOpndTypePtr DefOp, int DefSSANum, bool &Signed); // Is DefOp/DefAddr used in a loop back edge compare-and-branch?
	bool PropagateLocalDefType(STARSOpndTypePtr DefOp, SMPOperandType DefType, ea_t DefAddr, int SSANum, bool IsMemOp, bool PointerOverride = false); // to all uses
	bool InferLocalDefType(STARSOpndTypePtr DefOp, unsigned int LocIndex, ea_t DefAddr); // Can DEF type be inferred from all USEs?
	void InferGlobalDefType(STARSOpndTypePtr DefOp, int SSANum, ea_t DefAddr, bool &FoundNumeric, bool &FoundPointer,
		bool &FoundUnknown, bool &FoundUninit, SMPOperandType &PtrType);  // Can DEF type be inferred from all USEs? Recursive helper for SMPFunction::InferGlobal...
	bool PropagateGlobalDefType(STARSOpndTypePtr DefOp, SMPOperandType DefType, int SSANum, bool IsMemOp, bool PointerOverride, bool FirstBlock = true); // to all uses
	bool InferAllPhiDefTypes(void); // infer, propagate to all uses
	bool PropagateLocalMetadata(STARSOpndTypePtr UseOp, SMPMetadataType Status, int SSANum, ea_t UseAddr);
	bool FindRedundantLocalMetadata(bool SafeFunc); // consecutive DEFs, same type => redundant
	bool HasUseAfterIndWrite(STARSOpndTypePtr MemDefOp, std::vector<SMPInstr *>::iterator InstIter, bool &IndWriteSeen, bool &ReDef); // MemDefOp starting at InstIter has DU chain overlapping an indirect mem write
	void MarkBranchSignedness(void); // If branch at block end is signed/unsigned, propagate to operands that set flags before it.
	void PropagatePhiFGInfo(void); // Ensure that the Phi DEFs have all the fine grained info from all Phi USEs.
	void PropagateBranchSignedness(ea_t DefAddr, STARSOpndTypePtr SearchOp, unsigned short SignMask);
	bool PropagateDEFSignedness(void); // propagate DEF signedness to USE if USE has no signedness info.
	void MarkUnsignedArgs(ea_t CallAddr, unsigned int ArgPosBits); // backtrack from CallAddr and mark unsigned args based on bitset
	void MarkPointerArgs(ea_t CallAddr, unsigned int ArgPosBits); // backtrack from CallAddr and mark POINTER args based on bitset
	bool ComputeReachesInSet(void); // return true if changes were made to the ReachesInSet
	bool ComputeReachesOutSet(void); // return true if changes were made to the ReachesOutSet
	void UpdateDownExposedDefs(STARSOpndTypePtr DefOp, ea_t InstAddr); // Add DefOp, remove previous defs of DefOp that now do not reach the end of the block.

	void SCCPEvaluateConstants(enum STARSBranchConst &BranchEval, std::list<std::pair<int, int> > &CFGWorkList, std::list<std::pair<int, int> > &SSAWorkList); // Evaluate constants for all instructions as part of SCCP algorithm at SMPFunction level.
	void SCCPGlobalPropagationHelper(STARSOpndTypePtr GlobalOp, int DefSSANum, std::vector<STARSBitSet> ExecutedEdgeBitSet, std::list<std::pair<int, int> > &CFGWorkList, std::list<std::pair<int, int> > &SSAWorkList); // Propagate GlobalOp/DefSSANum const value
	void SCCPHandleSuccessors(enum STARSBranchConst &BranchEval, std::list<std::pair<int, int> > &CFGWorkList); // Based on BranchEval, push successors onto CFGWorkList
	void SCCPNullifyUnreachableBlock(void); // Patch binary to make this block into nops.
	void SCCPGatherStatistics(void); // Gather statistics on SCCP effectiveness

	// Find the stack target of a call to memset() and the size in bytes of the memset() region.
	//  If the memset() target is not on the stack, or the size is not a constant, return false.
	bool AnalyzeMemSet(ea_t MemSetAddr, STARSOpndTypePtr &MemSetTarget, size_t &MemSetSize, int &SignedOffset, bool &FPRelativeTarget);
	
	bool IsInfiniteSelfLoop(void); // return true if block loops infinitely to itself
	bool IsBenignOverflowDEF(STARSOpndTypePtr DefOp, int DefSSANum, size_t DefAddr, bool UnderflowOpcode, int &IdiomCode); // Do we not care if DEF overflowed, due to how it is used?
	bool IsDEFWrittenWithTruncation(STARSOpndTypePtr DefOp, int DefSSANum, ea_t DefAddr, ea_t TruncAddr, bool PhiDEF); // Is subword of DefOp written later, before TruncAddr?
	bool IsBenignTruncationDEF(STARSOpndTypePtr DefOp, int DefSSANum, size_t DefAddr, int &IdiomCode); // Do we not care if DEF overflowed, due to how it is used?
	bool IsDefMaskedOrShiftedRight(STARSOpndTypePtr DefOp, int DefSSANum, ea_t DefAddr); // Does DefOp+DefSSANum get subreg masked or right shifted before it gets re-DEFined?
	bool IsCriticalSink(STARSOpndTypePtr DefOp, int DefSSANum, std::string &SinkString, bool &FoundAnyCall); // What is ultimate use of DEF that might have integer error?
	bool IsStackOpNextUsedWithSignedness(STARSOpndTypePtr StackDefOp, ea_t DefAddr, int SSANum); // Is next use of stack DEF a sign- or zero-extended load?
	void SuppressAnnotOnSignChangingAddition(STARSOpndTypePtr DefOp, int DefSSANum, size_t DefAddr); // Find inc/add that makes small negative back into positive
	void AnalyzePrepForNumericAnnotations(void); // Last-minute prep before emitting numeric annotations
	// Is DefOp+DefSSANum (from WorkList head) at DefAddr used as address reg or as source operand in memory write?
	bool IsDefUsedInMemWrite(std::list<std::pair<std::pair<STARSOpndTypePtr, int>, ea_t> > &WorkList, STARSOpndTypePtr DefOp, int DefSSANum, ea_t DefAddr);

	sval_t ComputeStackAdjustmentAfterCall(ea_t CallAddr); // find stack pointer adjustment code after call, return stack delta or zero

private:
	// Data
	ea_t FirstAddr;
	int BlockNum;  // Number for block ordering algorithms
	int LoopHeaderNumber; // Number of block that is loop header for this block
	SMPFunction *MyFunc; // function containing this block
	sval_t OutgoingStackDelta; // Incoming stack ptr delta for all successor blocks
#if 0
	sval_t IncomingStackPtrDelta; // Incoming stack pointer delta for this block
#endif
	unsigned char booleans1; // hold bools as bits
#if 0  // now in booleans1
	bool Processed; // Set to true to prevent infinite recursions in CFG traversals

	// cached query results
	bool IndirectJump; // contains an indirect jump instruction
	bool Returns;      // contains a return instruction
	bool SharedTailChunk; // is part of a code chunk shared among functions
	bool IsLiveInStale; // Has LiveOutSet changed since LiveInSet was computed?
	bool FlagsDeadAfterLastUse; // flags global, dead after last use in the block
	bool FlagsDeadBeforeFirstKill; // flags global, dead before first DEF in block
	bool AliasedWrite; // A memory write in the block may be aliased by other writes
#endif
	unsigned char booleans2; // hold bools as bits
#if 0 // now in booleans2
	bool ReachesOutStale; // ReachesIn has been updated since last ReachesOut update
	bool LoopTailBlock; // block has back edge to loop header block
#endif

	std::vector<SMPInstr *> InstVec; //NOTE: Copies of pointers from SMPFunction::Instrs.
	std::map<ea_t, size_t> AddrInstMap; // map InstAddr to InstVec index.
	std::list<SMPBasicBlock *> Predecessors;
	std::list<SMPBasicBlock *> Successors;
		// Four sets used in live variable analysis
	std::set<STARSOpndTypePtr, LessOp> KillSet;   // variables killed in this block
	std::set<STARSOpndTypePtr, LessOp> UpExposedSet; // upward exposed variables in this block
	std::set<STARSOpndTypePtr, LessOp> LiveOutSet; // Live-Out variables in this block
	std::set<STARSOpndTypePtr, LessOp> LiveInSet; // contribution to predecessor's live-out iteration
		// Three sets used in reaching definitions analysis
	std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition> ReachesInSet; // definitions live on entry to block
	std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition> ReachesOutSet; // definitions live on exit from block, either passing through or DEFed inside block
	std::set<std::pair<STARSOpndTypePtr, ea_t>, LessDefinition> DownExposedDefnSet; // definitions within block that are live on exit
		// SSA data structures
	std::set<int> DomFrontier; // Dominance frontier for block, as set of RPO block numbers
	std::set<SMPPhiFunction, LessPhi> PhiFunctions; // SSA incoming edge phi functions
	std::set<STARSOpndTypePtr, LessOp> LocalNames; // non-global names referenced in this block
#if STARS_BUILD_DEF_USE_CHAINS
	SMPCompleteDUChains LocalDUChains; // def-use chains for local names
	SMPCompleteDUChains GlobalDUChains; // def-use chains for global names
		// NOTE: The GlobalChains.ChainsByName.at(GlobalIndex).DUChains are indexed
		//  starting at zero. The indices into the DUChains have nothing to do
		//  with the SSA Numbers, unlike the LocalDUChains.
#endif

	std::map<int, struct FineGrainedInfo> LocalDefFGInfoBySSA; // map hash of local name & SSANum to DEF FG info.
		// NOTE: We are currently limiting this map to registers, not all local names.

	std::map<int, struct FineGrainedInfo> LocalUseFGInfoBySSA; // map hash of local name & SSANum to USE FG info.
		// NOTE: We are currently limiting this map to registers, not all local names.

	std::map<int, struct STARS_SCCP_Const_Struct> ConstantLocalDefs; // map hash of local name & SSANum to constant value from SCCP.
		// NOTE: We are currently limiting this map to registers, not all local names. 

	// Methods
	inline void ResetReachesOutStale(void) { booleans2 &= BLOCK_RESET_REACHES_OUT_STALE; };
	inline void SetLoopTailBlock(void) { booleans2 |= BLOCK_SET_LOOP_TAIL_BLOCK; };
	std::set<SMPPhiFunction, LessPhi>::iterator InferPhiDefType(std::set<SMPPhiFunction, LessPhi>::iterator DefPhi, bool &changed); // infer, propagate to all uses
#if STARS_BUILD_DEF_USE_CHAINS
	unsigned int GetLocalDUIndex(STARSOpndTypePtr DefOp, int SSANum);
	unsigned int GetGlobalDUIndex(STARSOpndTypePtr DefOp, ea_t DefAddr);
#endif
	bool UseHasCriticalSink(STARSOpndTypePtr DefOp, int DefSSANum, STARSOpndTypePtr DefOp2, int DefSSANum2, std::string &SinkString, bool &FoundAnyCall); // DefOp or DefOp2 is passed as arg to critical system or lib call
	bool IsBranchSignednessUnreliable(void); // Is branch signedness misleading due to odd code generation idiom or hand coding?
}; 

#endif
